/**
 * @file   int_dict.c
 * @author David Bremner <bremner@mathematik.tu-muenchen.de>
 * @date   Thu Apr  3 20:03:57 2003
 * 
 * @brief  provide dictionaries (sparse, associative arrays) of integers
 * 
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include "int_dict.h"


int_dict_p id_new(int size){
  int_dict_p new;
  int i;

  if (size<=0){
    size=ID_DEFAULT_SIZE;
  }

  new=memalloc(sizeof(int_dict_t));

  new->table_size=size;
  new->table=calloc(size,sizeof(key_val_t));
  for (i=0; i< new->table_size; i++)
    new->table[i].key=-1;

  new->filled=0;
  new->probes=0;
  new->searches=0;
  new->next=NULL;
  return new;
}



static int hash(int k){
  int h=k*123143+23;

 if (h<0) 
   h=h*(-1);
  return h;
}


void id_destroy(int_dict_p dict){
  free(dict->table);
  free(dict);
}

void id_map(int_dict_p dict,id_hook func, void *arg){
  int i;
  for (i=0; i<dict->table_size; i++){
    if (dict->table[i].key>=0)
      func(arg,dict->table[i].key,dict->table[i].val);
  }
}

void id_flatten8(int_dict_p dict, int8_t *array){
  int i;
  for (i=0; i<dict->table_size; i++){
    if (dict->table[i].key>=0)
      array[dict->table[i].key]=dict->table[i].val;
  }
}

void id_flatten16(int_dict_p dict, int16_t *array){
  int i;
  for (i=0; i<dict->table_size; i++){
    if (dict->table[i].key>=0)
      array[dict->table[i].key]=dict->table[i].val;
  }
}

void id_flatten32(int_dict_p dict, int32_t *array){
  int i;
  int found=0;
  for (i=0; found<dict->filled && i<dict->table_size; i++){
    if (dict->table[i].key>=0){
      array[dict->table[i].key]=dict->table[i].val;
      found++;
    }
  }
}


void id_dump(int_dict_p dict){
  int i;
  char buf[1024]="";
  FILE *log;

  sprintf(buf,"int_dict_%x.log",dict);
  log=fopen(buf,"w");

  for (i=0; i<dict->table_size; i++){
    fprintf(log,"%d %d\n",dict->table[i].key,dict->table[i].val);
  }
  fclose(log);
}

void id_grow(int_dict_p dict){
  key_val_p old_table=dict->table;
  int old_size=dict->table_size;
  int i;

  dict->table_size *= ID_TABLE_MUL;
  dict->table_size += ID_TABLE_ADD;
  dict->table=memalloc(dict->table_size*sizeof(key_val_t));
  for (i=0; i< dict->table_size; i++){
    dict->table[i].key=-1;
  }
  
  for(i=0; i<old_size; i++){
    key_val_p pair=id_find(dict,old_table[i].key);

    pair->key=old_table[i].key;
    pair->val=old_table[i].val;
  }
  free(old_table);
}

int id_capacity(int_dict_p dict){
  return dict->table_size;
}

int id_filled(int_dict_p dict){
  return dict->filled;
}

int id_searches(int_dict_p dict){
  return dict->searches;
}

int id_probes(int_dict_p dict){
  return dict->probes;
}

#define SAFEDIV(a,b) ( b ? (double)a/b : 0)
int id_report(int_dict_p dict){
  fprintf(stderr,"dict %d %d %d %d\n",
	  dict->filled,dict->table_size,
	  dict->searches,dict->probes);
}

key_val_p id_find(int_dict_p dict, int key){

  int h;

  int stop;

  h= hash(key) % (dict->table_size);
  stop=h;

  dict->probes++;

  while (dict->table[h].key >= 0  &&
	 dict->table[h].key != key){
    h = (h+1) % dict->table_size;
    dict->probes++;
  }
  return dict->table+h;
}
  

void id_set(int_dict_p dict,int key, int val,bool overwrite){
  key_val_p pair;

  while((double)dict->filled/(dict->table_size)>ID_FILL_LIMIT)
    id_grow(dict);

  dict->searches++;
  pair=id_find(dict,key);
  if (pair->key==-1){
    dict->filled++;
    pair->key=key;
  } else {
    if (!overwrite)
      return;
  }


  pair->val=val;
}
  
