/**
 * @file   matroid.c
 * @author David Bremner <bremner@unb.ca>
 * @date   Mon Dec 30 18:21:28 2002
 * 
 * @brief  enforce the basis exchange axiom
 * 
 */

#include "nalloc.h"
#include "basis_equation.h"
#include "combinat.h"
#include "globaldef.h"
#include "nodedef.h"
#include "bitmask.h"
#include "error_message.h"
/** 
 * allocate a block for matroid (basis-exchange) equations.
 * 
 * @param state node to hold block
 * @param n elements
 * @param r rank
 * 
 * @return modified node
 */
na_node_p me_nalloc(na_node_p state,slot_t slot, int  n, int r){
  int rows=binomial(n,r)*r;
  int cols=r+1;

  state=be_nalloc(state,slot,rows,cols,binomial(n,r));
  
  return state;
}
static void pivot(int r, 
		  int *basis,
		  int exit,
		  int enter,
		  int *output){

  int i;

  for (i=0; i< r; i++){
    if (basis[i] == exit)
      output[i]=enter;
    else
      output[i]=basis[i];
  }
}
		  
		  

/** 
 * initialize a set of constraints for checking basis exchange axiom
 * to understand this routine, see Ex 3.21 in Bjorner et al.
 * @param matroid_eq preallocated block
 * @param n elements
 * @param r rank
 * @param beta0 "root" basis, guaranteed nonzero by caller
 *
 */
void me_init(be_block_p matroid_eq, sarray_p zero_mask,
	     int n, int r,array_index_t beta0){

  natural_t *pi,*root_pi,*enterers,*leavers,*neighbour;
  int row=0;

  matroid_eq->ones_mask= (1 << matroid_eq->cols)-1;

  root_pi=calloc(r,sizeof(natural_t));
  pi=calloc(r,sizeof(natural_t));
  enterers=calloc(r+1,sizeof(natural_t));
  leavers=calloc(r+1,sizeof(natural_t));
  neighbour=calloc(r+1,sizeof(natural_t));
 
  which_kset(n,r,beta0,root_pi);

  first_kset(n,r,pi);

  /* iterate over bases */
  do {
    int l,e;
    int ecount,lcount;

    /* find B_0 \B' */
    lcount=kset_diff(r,root_pi,pi,leavers);

    /* find B' \ B_0 */
    ecount=kset_diff(r,pi,root_pi,enterers);
    
    /* iterate over leaving vertices */

    for (l=0; l< lcount; l++){
      bitmask_t initial_mask=matroid_eq->ones_mask<<1;

      /* save the 'leaf' basis */
      be_set_bindex(matroid_eq,row,0,kset_index(n,r,pi));
      be_set_sign(matroid_eq,row,0,1);

      for (e=0; e<ecount; e++){
	pivot(r,root_pi,leavers[l],enterers[e],neighbour);
	be_set_bindex(matroid_eq,row,e+1,kset_cindex(n,r,neighbour));
	be_set_sign(matroid_eq,row,e+1,1);
	initial_mask <<= 1;
      }
      sa_init(zero_mask,row,initial_mask&matroid_eq->ones_mask);
      for (e=ecount; e< r; e++){
	be_mark_unused(matroid_eq,row,e+1);
      }
      row++;
    }
     
  } while(next_kset(n,r,pi));


  free(pi);
  free(root_pi);
  free(enterers);
  free(leavers);
  free(neighbour);

  /* note, this is really a gross use of space */
  for (/*nothing*/; row<be_rows(matroid_eq); row++){
    int j;
    for (j=0; j<be_cols(matroid_eq); j++){
      be_mark_unused(matroid_eq,row,j);
    }
  }

  be_makelinks(matroid_eq);
  matroid_eq->initialized=1;
}


void me_setup(na_node_p global,na_node_p local){

    sarray_p chi=na_access(local,SLOT_chirotope);
    params_p params=na_access(global,SLOT_params);
    sarray_p nonzero_mask=na_access(local,SLOT_matroid_nonzero_mask);

    int i;

    
    for (i=0; i<sa_size(nonzero_mask); i++)
      sa_init(nonzero_mask,i,0);
    
    
}

