#include "basis_equation.h"
#include "bitmask.h"
#include "boundary_state.h"
#include "boundary_state.h"
#include "combinat.h"
#include "debug.h"
#include "facet_list.h"
#include "globaldef.h"
#include "nodedef.h"
#include "nu_sign_set.h"
#include "oms_opt.h"
#include "profile.h"

static bool force_row(be_block_p BE, sarray_p chi, array_index_t bindex, array_index_t elem, nu_sign_t sigma, nu_sign_set_p forced);


static bool nb_process_row(na_node_p global, na_node_p local, 
			   be_block_p BE, sarray_p chi, 
			   array_index_t elem, nu_sign_t sigma,
			   bitmask_t pmask, bitmask_t nmask, bitmask_t zmask, 
			   nu_sign_set_p forced, oms_hook force_hook);




/** 
 * force row
 * 
 * @param BE boundary equations block
 * @param elem pseudopointer to cell
 * @param sigma sign to force row containing elem
 * 
 * @return false if contradiction encountered
 */

static bool force_row(be_block_p BE,
		      sarray_p chi,
		      array_index_t bindex,
		      array_index_t elem, 
		      nu_sign_t sigma,
		      nu_sign_set_p forced){
  int i;

  int row_start=be_row_start(BE,elem);
  for (i=row_start; 
       i<row_start+BE->cols && ( be_in_use(BE,i)); i++){
    array_index_t b=be_bindex_raw(BE,i);
    nu_sign_t chi_i;
    nu_sign_t weak_sigma=sign_union(sigma,SIGN_ZERO);

    err_assert(b >=-1 && b <= sa_size(chi));
    if (b != bindex){
	      
      chi_i=sign_mulint(sa_read(chi,b),be_sign_raw(BE,i));
      if (!sign_subset(chi_i,weak_sigma)){
	if (!sign_compatible(weak_sigma,chi_i)){
	  ERR_RETURN(FALSE);
	} else {
	  nu_sign_t new=sign_mulint(sign_intersection(weak_sigma,chi_i),
				    be_sign_raw(BE,i));
	  err_debug(DB_BOUNDARY|DB_FORCE,(HERE, "row %d forcing chi[%d] to %s",
					 be_row(BE,i),b,sign_name[new]));
	  if (!nss_add(forced,b,new)) ERR_RETURN(FALSE);
	} 
      }
    }
  }
  ERR_RETURN(TRUE);
}

/** 
 * Traverse boundary equations.

 * This version provides a hook to be called for each forced facet.
 * It requires a three more state arrays, pos_mask,  neg_mask, and 
 * zero mask to 
 * be defined.


@param  bindex  what basis to check
@param	local	backtracking search state
@param	global	global variables (read only)
@param	forced	Queue of basis signs forced by this traversal
@param	force_hook function to call for each simplex forced onto the boundary.
*/
bool nb_traverse(array_index_t bindex,
		 na_node_p local, 
		 na_node_p global,
		 nu_sign_set_p forced,
		 oms_hook force_hook) {
  array_index_t elem;
  nu_sign_t chi_bindex;
  sarray_p chi, pos_mask,neg_mask, zero_mask;
  be_block_p BE;

  chi=na_access(local,SLOT_chirotope);
  BE=na_access(global,SLOT_bdy_eq);
  pos_mask=na_access(local,SLOT_pos_mask);
  neg_mask=na_access(local,SLOT_neg_mask);
  zero_mask=na_access(local,SLOT_zero_mask);

  chi_bindex=sa_read(chi,bindex);
  elem=BE->last[bindex];   
  err_assert( elem >= -1 && elem < BE->rows*BE->cols );
  while ( elem>=0  ){
    array_index_t row,col;
    int32_t pmask,nmask,zmask;
    nu_sign_t sigma;


    row=be_row(BE,elem);
    col=be_col(BE,elem);

    sigma=sign_mulint(chi_bindex,be_sign_raw(BE,elem));

    
    pmask=be_update_mask(pos_mask,SIGN_POS,row,col,sigma);
    nmask=be_update_mask(neg_mask,SIGN_NEG,row,col,sigma);
    zmask=be_update_mask(zero_mask,SIGN_ZERO,row,col,sigma);

    if (ERR_DEBUG_MASK & (DB_BOUNDARY|DB_MASKS)){
      be_check_mask(global,BE,row,pmask,SIGN_POS,chi);
      be_check_mask(global,BE,row,nmask,SIGN_NEG,chi);
      be_check_mask(global,BE,row,zmask,SIGN_ZERO,chi);
    }

    if (!nb_process_row(global,local,BE,chi,
			elem,sigma,pmask,nmask,zmask,forced,force_hook))
      return(FALSE);
    
      
    elem=BE->eqn[elem].prev;  
  } /* while (elem>0) */

  ERR_RETURN(TRUE);
}

#if 0
static bool nb_facet_orientation(array_index_t row, nu_sign_t sigma,
				 na_node_p global, na_node_p local){

  array_index_t *facet_id=na_access(global,SLOT_facet_id);
  sarray_p facet_orientation=na_access(local,SLOT_facet_orientation);

  array_index_t fid=facet_id[row];

  nu_sign_t cur_sign;

  return TRUE; /* Apparently this check is wrong, and uneccesary
		See Lemma 3.5.8 in Bjorner et al.*/
  /* if this simplex has not yet been assigned to the boundary, 
     there is nothing we can say */
  if (fid<0)
    ERR_RETURN(TRUE);

  /* if this basis might be flat, then it tells us nothing definite
     about which way the facet is oriented */
  if (sign_subset(SIGN_ZERO,sigma))
    ERR_RETURN(TRUE);

  cur_sign=sa_read(facet_orientation,fid);

  if (!sign_compatible(cur_sign,sigma)){
    err_debug(DB_BOUNDARY|DB_ORIENTATION,
	      (HERE,"orientation conflict facet=%d, row=%d,"
	       "new=%s current=%s",fid,row,
	       sign_name[sigma],sign_name[cur_sign]));
    ERR_RETURN(FALSE);
  }

  if (!sign_subset(cur_sign,sigma)){
    err_debug(DB_BOUNDARY|DB_ORIENTATION,(HERE,"row %d orienting facet %d to %s\n",
			   row,fid,sign_name[sigma]));

    sa_set(facet_orientation,fid,sigma);
  }
  ERR_RETURN(TRUE);
}

#endif

static bool nb_process_row(na_node_p global,
		    na_node_p local,
		    be_block_p BE,
		    sarray_p chi,
		    array_index_t elem,
		    nu_sign_t sigma,
		    bitmask_t pmask, 
		    bitmask_t nmask, 
		    bitmask_t zmask,
		    nu_sign_set_p forced,
		    oms_hook force_hook){



  ridge_graph_p RG=na_access(local,SLOT_ridge_graph);
  be_block_p pivots=na_access(global,SLOT_pivots);
  sarray_p active=na_access(local,SLOT_boundary);

  bdy_state_t row_active;

  array_index_t row=be_row(BE,elem);
  array_index_t bindex=be_bindex_raw(BE,elem);
  
  row_active=sa_read(active,row);


  err_debug(DB_BOUNDARY,
	    (HERE,"row=%d, active=%d, nmask=%x zmask=%x pmask=%x",
	       row,row_active, nmask,zmask,pmask));

  switch (row_active){
  case BDY_MAYBE:
    /* Currently Undecided */
    if ( (pmask|nmask|zmask) == BE->row_mask[row] ){
	if ((pmask==0) || (nmask==0)){
	  err_debug(DB_BOUNDARY|DB_FORCE,
		    (HERE, "forcing %d onto boundary",row));
	  if (OPT_use_profile){
	    prof_event(PROF_BOUNDARY_FORCED);
	  }
		
	  sa_set(active,row,BDY_YES);
	  /* \todo, some things should be forced to zero here */
	  /* namely, everything in (pmask&nmask)&zmask */

	  if (force_hook && !force_hook(local,global,row,BDY_YES))
	    ERR_RETURN(FALSE);
	  
	  if (BE->diameter !=-1){
	    if (!rg_add_facet(RG,row,active,pivots))
	      ERR_RETURN(FALSE);
	  }
	} else {
	  err_debug(DB_BOUNDARY|DB_FORCE,(HERE, "forcing %d off boundary",row));

	  sa_set(active,row,BDY_NO);
	}
      }
      break;
      
    case BDY_YES:
      
      /* if there are vertices on both sides, fail */
      if ((pmask  & ~(zmask|nmask))  && ( nmask & ~(zmask|pmask)))
	ERR_RETURN(FALSE);

      /* if this facet generates both definitely flat, and definitely
	 non-flat bases, fail */
      if ( (zmask & ~(pmask|nmask))   && ((pmask|nmask) & ~zmask)){
	ERR_RETURN(FALSE);
      }
      
      /* try and force all of the bases in this row to match */
      if (sigma != SIGN_ZERO && (pmask|nmask|zmask) != BE->row_mask[row]){
	  if (!force_row(BE,chi,bindex,elem,sigma,forced)) ERR_RETURN(FALSE);
      }

      break;

    case BDY_NO:      /* Not permitted on the boundary */
      if (!be_nonmonotone_row(BE,elem,sigma,pmask,nmask,zmask,forced))
	ERR_RETURN(FALSE);
	   
      break;
    default:
      err_fatal(HERE,"Unknown row_active value %d",row_active);
    }/* switch (row_active) */
  ERR_RETURN(TRUE);
}

/*
 * initialize non-uniform boundary equations.
 * We still work on an (r-1)-simplex by (r-1)-simplex 
 * The main difference is that we rely on e.g. Lemma 3.5.8 
 * of the red book, which tells us that we don't need to worry
 * about matching the tests of different simplices from the same 
 * facet.
 *
 * Also, we trim the extension lists to consider only extensions 
 * off of the current facet.
 */
extern void nb_boundary_init(be_block_p bdy_eq, natural_t n, natural_t r,
			     facet_list_p fl, array_index_t source, array_index_t sink,
			     array_index_t diameter){

  natural_t num_bases,num_eqns;

  facet_p *facet;
  natural_t *pi,*pibar;
  int *tau;
  int d=r-1;
  int row;
  facet_p F;

  num_bases=binomial(n,r);
  num_eqns=binomial(n,r-1);

  err_assert(bdy_eq->cols==n-r+1);
  err_assert(bdy_eq->bases==num_bases);
  err_assert(bdy_eq->rows==num_eqns);
  err_assert(bdy_eq->eqn !=NULL);

  bdy_eq->source=source;
  bdy_eq->sink=sink;
  bdy_eq->diameter=diameter;

  facet=calloc(fl_length(fl),sizeof(facet_p));

  for (F=fl_first(fl); F; F=fl_next(F)){
    facet[F->facet_id]=F;
  }

  bdy_eq->stride=1;
  bdy_eq->source=-1;
  bdy_eq->sink=-1;
  bdy_eq->diameter=-1;


  pi=calloc(d+1,sizeof(natural_t));
  pibar=calloc(n-d,sizeof(natural_t));
  tau=calloc(n-d,sizeof(int));

  row=0;
  first_kpartition(n,d,pi,pibar,tau);
  
  do{

    int j;
    int col=0;
    bitmask_t rmask=~0U;

    array_index_t fbindex=kset_cindex(n,d,pi);
    array_index_t fid=fl_facet_id(fl,fbindex);
    
    facet_p F;
    
    if (fid>=0)
      F=facet[fid];

    for (j=0; j< n-d; j++){
      if (fid<0 || !set_contains(F->size,F->index_set,pibar[j])){
	be_set_sign(bdy_eq,row,col,tau[j]);
	swap(pi[d],pibar[j]);
	be_set_bindex(bdy_eq,row,col,kset_cindex(n,r,pi));
	swap(pi[d],pibar[j]);
	col++;
	rmask <<= 1;
      }
    }
    bdy_eq->row_mask[row]=~rmask;

    for (j=col; j<n-d; j++){
      be_set_sign(bdy_eq,row,j,-1);
      be_set_bindex(bdy_eq,row,j,-1);
    }
    row++;
  } while(next_kpartition(n,d,pi,pibar,tau));

  free(pi);
  free(pibar);
  free(tau);

  be_makelinks(bdy_eq);  
}

