/* z_base.c (ZRAM: base definitions)
   Ambros Marzetta

 * $Source$
 * $Revision: 3146 $
 * $Date: 2004-08-31 02:08:56 +0200 (Tue, 31 Aug 2004) $
 * $Log$
 * Revision 1.14  2004/08/31 00:08:56  bremner
 * working on standalone install
 *
 * Revision 1.13  2004/08/25 21:01:45  bremner
 * libutil configuration
 * zram without mpi
 *
 * Revision 1.12  2004/06/22 21:02:50  bremner
 *
 * in the middle of debugging memory problems in zram mode
 *
 * Revision 1.11  2002/07/22 15:43:45  bremner
 * *** empty log message ***
 *
 * Revision 1.10  2002/07/18 15:21:22  bremner
 * *** empty log message ***
 *
 * Revision 1.9  2002/04/04 15:55:07  bremner
 * *** empty log message ***
 *
 * Revision 1.8  2002/04/04 14:34:47  bremner
 * about to rip out a bunch of custom memory allocation code and replace
 * it with calls to memalloc. Because they are making me crazy
 *
 * Revision 1.7  2002/04/03 11:50:53  bremner
 * tracking ugly memory leaks
 *
 * Revision 1.6  2002/04/02 15:28:39  bremner
 * Fixed bug with forgetting to forget pos_mask and neg_mask.
 *
 * Revision 1.5  2001/08/09 20:51:46  bremner
 * malloc fun continues.  Leakage mostly staunched for short1.om
 *
 * Revision 1.4  2001/08/08 21:07:41  bremner
 * beginnings of a configuration system. Mostly I am bored with creating
 * new revisions of Makefile
 *
 * Revision 1.3  2001/08/06 18:02:04  bremner
 * changing nalloc memory management. Not currently stable
 *
 * Revision 1.2  2000/11/15 16:14:52  bremner
 *
 * committing patched version of zram source
 *
 * Revision 1.8  1996/06/03  19:00:37  marzetta
 * New definition of z_NewNode.
 *
 * Revision 1.7  1996/05/29  11:32:56  marzetta
 * eliminated compiler warnings
 *
 * Revision 1.6  1996/04/09  13:20:21  marzetta
 * new Install procedures
 *
 * Revision 1.5  1995/09/14  12:25:56  marzetta
 * deep copy
 *
 * Revision 1.4  1995/08/22  12:39:12  marzetta
 * New message passing, MPI
 *
 */

#include <stdio.h>
#include <stdlib.h> /* exit() */
#include <string.h> /* memcpy() */
#include "z_base.h"
#ifdef HAVE_CONFIG_H
#include "zram_config.h"
#endif

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "zram_debug.h"
#include "z_memalloc.h"

void z_DumpMemUsage();
/* Interrupts ----------------------------------------------------------- */

volatile boolean z_terminate;
/* Will be set when the process receives a terminate signal. */

volatile boolean z_checkpoint;
/* Will be set when some timer expires. */

/* Installing classes --------------------------------------------------- */

#define MAXCLASS Z_NUM_TAGS
#define NUMCLASSSTART 20	/* should be 1 */

static int numClass = NUMCLASSSTART;
static const char *className[MAXCLASS];
static int classChildren[MAXCLASS];

int z__InstallClass(const char *name, int children,
		    int reserved1, int reserved2)
/* Install a node type. z__InstallClass returns a tag to be used in
   subsequent calls to z_NewNode and in accesses to z_numNodes. This is
   a collective operation. */
{
  assert(reserved1 == 0);
  assert(reserved2 == 0);
  assert(numClass < MAXCLASS);
  className[numClass] = name;
  classChildren[numClass] = children;
  /*printf("Installing %s with tag %d and %d children\n", name, numClass, children);*/
  return numClass ++;
} /* z_InstallClass */

/* Memory management ---------------------------------------------------- */

#define FREELISTSIZE 10

typedef struct z_FamilyNode {
  Z_NODE_HEADER;
  z_NodeP ch[10];
} *z_FamilyNodeP;

long z_numNodes[Z_NUM_TAGS];

#if 0
static struct z_Node freeList;
int freeSize;
#endif

static z_FamilyNodeP z_Allocate(MINT len)
{
  z_NodeP f;
  z_FamilyNodeP new;
  z_memavail -= len;

#ifdef ZRAM_MEMORY_LIMIT
  if (z_memavail < 0){
    z_DumpMemUsage();
#ifdef USE_DMALLOC
    dmalloc_shutdown();
#endif
    abort();
  }
#endif
  new=memalloc(len);
  return new;
} /* z_Allocate */


void z_DumpMemUsage(){
    int i;
    
    printf("Allocated nodes:\n");
    for (i=NUMCLASSSTART; i<numClass; i++)
      printf("  %-20s %4ld\n", className[i], z_numNodes[i]);
}

z_NodeP z_NewNode(MINT tag, MINT len)
/* Allocate a z_Node, initialize tag, len, children */
{
  z_FamilyNodeP np;

  assert(0 < tag && tag < Z_NUM_TAGS);
  /* forbid tag == 0 to detect uninitialized tags */
  assert(className[tag]);
  /*printf("Allocating %ld bytes (tag = %ld)\n", (long)len, (long)tag);*/
  np = z_Allocate(len);
  if (np) {
    np->len = len;
    np->tag = tag;
    np->children = classChildren[tag];
    np->parent = NULL;
    np->child = NULL;
    np->sibling = NULL;
    z_numNodes[tag] ++;
  }
  if (! np) {			/* where is the memory lost? */
    z_DumpMemUsage();
  }
  return (z_NodeP) np;
} /* z_NewNode */

/* 
   ZRAM equivalent of realloc. Added by db.
   NOTE: for consistency, if not convenience, 
   newlen includes the ZRAM header size.  
*/


z_NodeP z_ReallocNode(z_NodeP source, MINT newlen){

  if (source->len >= newlen) 
    return source;

  z_memavail -= (newlen-source->len);
  
  source=realloc(source,newlen);

  source->len=newlen;
   
  return(source);
  
}

z_NodeP z_GrowNode(z_NodeP src, MINT bytes){
  return z_ReallocNode(src,src->len+bytes);
}

z_NodeP z_NewCopyNode(z_NodeP source)
{
  /* Allocate a z_Node identical to source (deep copy) */
  z_FamilyNodeP src, dest;
  int i;
  
  assert(source);
  src = (z_FamilyNodeP) source;
  dest = z_Allocate(source->len);
  if (! dest) {			/* where is the memory lost? */
    int i;

    
    printf("Allocated nodes:\n");
    for (i=NUMCLASSSTART; i<numClass; i++)
      printf("  %-20s %4ld\n", className[i], z_numNodes[i]);
#if 0
    printf("Free list:\n");
    for (f = freeList.child; f; f = f->child)
      printf("  %-20s      %d\n", className[f->tag], f->len);
#endif
  }

  if (dest) {
    z_numNodes[source->tag] ++;
    memcpy((char *) dest, (char *) source, source->len);
    for (i=src->children-1; i>=0; --i)
      if (src->ch[i]) dest->ch[i] = z_NewCopyNode(src->ch[i]);
  } else {
    printf("z_NewCopyNode: no memory (%ld nodes allocated)\n",
	   z_numNodes[source->tag]);
    fflush(stdout);
  }
  assert(dest);
  return (z_NodeP) dest;
} /* z_NewCopyNode */

void z_FreeNode(z_NodeP np)
{
  z_FamilyNodeP f;
  int i;
  
  err_debug(DB_MEMORY,(HERE,"Freeing %x\n",np));
  f = (z_FamilyNodeP) np;
  for (i=f->children-1; i>=0; --i)
    if (f->ch[i]) z_FreeNode(f->ch[i]);
  
  z_numNodes[np->tag] --;
  z_memavail += np->len;
  memfree(np);			/* z_FreeNode */
} /* z_FreeNode */

void z_ReSizePointer(void **ptr, long *size, long newsize)
{
  /* Allocate if *size < newsize. Don't copy any data. */
  if (newsize > *size) {
    if (*ptr)
      (void) free(*ptr);	/* z_ReSizePointer */
    *size = newsize;
    *ptr = (void *) malloc(newsize); /* z_ReSizePointer */
    assert(*ptr);
  }
} /* z_ReSizePointer */

