/* z_par.h (parallel base functions)

   Ambros Marzetta
   94-11-15

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 17:14:52 +0100 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.7  1996/04/09  13:27:12  marzetta
 * Added SEQUENTIAL, Cenju-3, new install procedures
 *
 * Revision 1.6  1995/09/14  12:27:52  marzetta
 * AddDataBuffer, GetDataBuffer
 *
 * Revision 1.5  1995/08/22  12:39:45  marzetta
 * New message passing, MPI
 *
 * Revision 1.3  1995/08/03  08:56:13  marzetta
 * Macintosh portability
 *
 * Revision 1.2  1995/07/27  11:36:00  marzetta
 * Separation of message passing interface
 *
 */

#ifndef _Z_PAR_H_
#define _Z_PAR_H_
#include "z_base.h"

/* Processor numbering ------------------------------------------------ */

#define ROOT_PE 0 /* A processor with this number always exists. */

#define MAX_PE 150		/* z_numprocs <= MAX_PE */

extern int z_numprocs; /* read only */
/* Number of processing elements. They are numbered from 0..z_numprocs-1.
   Processor number 0 always exists.
   There may be an additional processor (front end) with number -1.
   */

extern int z_myproc; /* read only */
/* A number from -1..z_numprocs-1. */

extern boolean z_frontend; /* read only */
/* Does this process run on the front end (Music) or on processor 0 (Paragon)?
   z_frontend returns true on exactly one processor. If there is a processor
   with number -1, it is the front end. Otherwise processor 0 is front end and
   processing element. (see z_WaitWork()) */

/* Buffers ----------------------------------------------------------- */

typedef int32 z_tagT;

#ifndef SEQUENTIAL
extern long z_bufsize; /* read only */
/* Read only. Size of a buffer (excluding the header).
   Do not add more data to the output buffer! */

/* The output buffer is emptied
   - at initialisation
   - in z_StartParallelJob
   - at the end of the execution of a job
   - after z_Isend and z_Ibroadcast when z_TestSend returns true
   */

void z_AddTagBuffer(z_tagT tag);
/* output buffer */

void z_AddMINTBuffer(MINT x);
/* output buffer */

void z_AddNodeBuffer(z_NodeP source);
/* output buffer; source may be NULL */

void z_AddDataBuffer(const void *source, long length);

/* The input buffer is full
   - when an installed job is called
   - after the execution of z_WaitWork
   - when a resultProc is called by z_WaitWork
   - when z_GotMessage returns true
   - when z_Idle returns false
   - when an installed message handler is called
    */

z_tagT 	z_GetTagBuffer(void);
MINT 	z_GetMINTBuffer(void);
z_NodeP	z_GetNodeBuffer(void);
void 	z_GetDataBuffer(void *dest, long length);

/* Starting jobs ---------------------------------------------------- */

typedef uint16 z_JobNum;

typedef void (* z_ResultProc)(int pe); /* pe = number of sending pe */

#define z_InstallJob(job) z__InstallJob(job, #job)

z_JobNum z__InstallJob(const z_VoidProc job, const char *name);

/* Jobs are installed e.g. by z_bb1_Init(), which is called by initproc().
   See z_StartParallelJob().
   */

#ifndef PEonly

void z_StartParallelJob(z_JobNum jobnum, z_ResultProc resultProc);
/* (FE only) Broadcast outbuffer and initiate the parallel computation.
   All PEs will call the installed job with number jobnum.
   z_StartParallelJob overwrites the tag of outbuffer.
   The job must first read the data in inbuffer. Then it can communicate
   using one of the following:
   - ZRAM message passing with termination detection (z_mp.h)
                                                (portable, except on music)
   - MPI message passing (mpi.h)		(portable, except on music)
   - Paragon message passing (nx.h)		(not portable)
   - Music communication (musiclib.h)		(not portable)
   */

void z_WaitWork(void);
/* (FE only) Let the job initiated by z_StartParallelJob() finish.
   On computers with a processor numbered -1 (Music frontend) this
   procedure simply waits for the results to arrive.
   On other computers (frontend has processor number 0: Paragon,
   sequential workstation) the frontend works as a processing element.

   If resultProc == NULL:
   At the end the job on PE 0 must place the result of the computation
   in outbuffer. ZRAM will transfer this buffer to the front end. After
   z_WaitWork returns, the message can be read from inbuffer.

   If resultProc != NULL:
   At the end every PE can place the result of its computation in
   outbuffer. ZRAM will transfer these buffers to the front end and call
   ResultProc() on the front end once for every PE. ResultProc should
   read the message from inbuffer. The messages from the different PEs
   may arrive in any order.
   */

#endif /* ! PEonly */
#endif /* SEQUENTIAL */

/* Assert ---------------------------------------------------------- */

/* Parallel assert. Must be included after <assert.h>, i.e. after "z_base.h"
   because it redefines the macro assert() */

void z_assert(const int, const char *, const char *, const int);

#undef assert
#ifdef NDEBUG			/* no debug version */
#define assert(ignore) ((void)0)
#else /* NDEBUG */		/* debug version */
#define assert(value) ((value)? (void)0 : z_assert(z_myproc, #value,__FILE__,__LINE__))
#endif /* NDEBUG */

#endif /* _Z_PAR_H_ */
